from eyelab.dialogs.help.help_dialog import HelpWindow

help_text = """
# Introduction

EyeLab is a multimodal annotation tool for ophthalmological imaging data.

Get started by importing data in one of the supported file formats

## Currently supported file formats

+ Data exported from HEXEX in the .xml or .vol format
+ Data from the RETOUCH challenge
+ AMD and Control Dataset from  (Farsiu 2014)
+ Import B-scans from folder (sorting the file names is expected to result in the correct B-scan order
+ Saving and opening of .eye files, EyeLabs file format.

## The Workspace:

Your workspace consists of a combined Enface and Volume view. On the right you have an overview
where annotations are managed for the OCT volume as well as for the Enface image.

## Navigation

You can zoom in and out using the mouse wheel. With the standard **Inspection** tool you can
move the image around if you want to see another part. If you press CTRL while using the mouse
wheel on the Volume view, you scroll through adjacent B-scans. Use CRTL + x to toggle a
linked navigation where B-scans change based on your mouse position in the enface.

## How to continue
From here you can continue with one of the following guides depending on your
needs.
+ Layer Annotation Guide
+ Area Annotation Guide
"""


class IntroductionHelp(HelpWindow):
    def __init__(self, parent):
        super().__init__(help_text, parent)



from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtWidgets import QWidget

from eyelab.views.ui.ui_layer_entry import Ui_LayerEntry
from eyelab.views.ui.ui_layergroup_entry import Ui_LayerGroupEntry


class LayerGroupEntry(QWidget, Ui_LayerGroupEntry):
    editorChanged = QtCore.Signal()

    def __init__(self, parent):
        super().__init__(parent)
        self.setupUi(self)

        self.visible_icon = QtGui.QIcon()
        self.visible_icon.addPixmap(
            QtGui.QPixmap(":/icons/icons/baseline-visibility-24px.svg"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        self.hidden_icon = QtGui.QIcon()
        self.hidden_icon.addPixmap(
            QtGui.QPixmap(":/icons/icons/baseline-visibility_off-24px.svg"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )

        self.visible = True
        self.hideButton.clicked.connect(self.toggle_visibility)

    def set_visible(self, value=True):
        if value:
            self.show()
        else:
            self.hide()

    def toggle_visibility(self):
        if self.visible:
            self.hide()
        else:
            self.show()
        self.editorChanged.emit()

    def hide(self):
        self.visible = False
        self.hideButton.setIcon(self.hidden_icon)

    def show(self):
        self.visible = True
        self.hideButton.setIcon(self.visible_icon)


class LayerEntry(QWidget, Ui_LayerEntry):
    editorChanged = QtCore.Signal()

    def __init__(self, parent):
        super().__init__(parent)
        self.setupUi(self)

        self.visible_icon = QtGui.QIcon()
        self.visible_icon.addPixmap(
            QtGui.QPixmap(":/icons/icons/baseline-visibility-24px.svg"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        self.hidden_icon = QtGui.QIcon()
        self.hidden_icon.addPixmap(
            QtGui.QPixmap(":/icons/icons/baseline-visibility_off-24px.svg"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )

        self.visible = True
        self.hideButton.clicked.connect(self.toggle_visibility)
        self.colorButton.clicked.connect(self.set_color)
        self.colorDialog = QtWidgets.QColorDialog()

        self.label.mouseDoubleClickEvent = self._enable_label_edit_event

        self.labelEdit = QtWidgets.QLineEdit()
        self.labelEdit.editingFinished.connect(self._disable_label_edit)
        self.labelEdit.hide()
        self.horizontalLayout.addWidget(self.labelEdit)

    def _enable_label_edit_event(self, event):
        self.labelEdit.setText(self.label.text())
        self.label.hide()
        self.labelEdit.show()

    def _disable_label_edit(self):
        self.label.setText(self.labelEdit.text())
        self.labelEdit.hide()
        self.label.show()
        self.editorChanged.emit()

    def leaveEvent(self, event: QtCore.QEvent) -> None:
        if self.labelEdit.isVisible():
            self._disable_label_edit()

    def set_color(self, color=None):
        if not color:
            color = self.colorDialog.getColor(
                options=QtWidgets.QColorDialog.DontUseNativeDialog
            )
            if not color.isValid():
                return
            self.color = color.name()[1:]
        else:
            self.color = color
        self.colorButton.setStyleSheet(f"background-color: #{self.color}")
        self.editorChanged.emit()

    def set_visible(self, value=True):
        if value:
            self.show()
        else:
            self.hide()

    def toggle_visibility(self):
        if self.visible:
            self.hide()
        else:
            self.show()
        self.editorChanged.emit()

    def hide(self):
        self.visible = False
        self.hideButton.setIcon(self.hidden_icon)

    def show(self):
        self.visible = True
        self.hideButton.setIcon(self.visible_icon)

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent) -> None:
        self.labelEdit.setFrame(True)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_scene_tab.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore

from . import resources_rc


class Ui_SceneTab(object):
    def setupUi(self, SceneTab):
        if not SceneTab.objectName():
            SceneTab.setObjectName(u"SceneTab")
        SceneTab.resize(166, 386)
        self.verticalLayout = QVBoxLayout(SceneTab)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.opacityLayout = QHBoxLayout()
        self.opacityLayout.setObjectName(u"opacityLayout")
        self.opacitySliderLabel = QLabel(SceneTab)
        self.opacitySliderLabel.setObjectName(u"opacitySliderLabel")
        font = QFont()
        font.setPointSize(7)
        font.setBold(True)
        self.opacitySliderLabel.setFont(font)

        self.opacityLayout.addWidget(self.opacitySliderLabel)

        self.opacitySlider = QSlider(SceneTab)
        self.opacitySlider.setObjectName(u"opacitySlider")
        self.opacitySlider.setMaximum(100)
        self.opacitySlider.setPageStep(10)
        self.opacitySlider.setValue(100)
        self.opacitySlider.setSliderPosition(100)
        self.opacitySlider.setOrientation(Qt.Horizontal)
        self.opacitySlider.setInvertedAppearance(False)
        self.opacitySlider.setInvertedControls(False)

        self.opacityLayout.addWidget(self.opacitySlider)

        self.verticalLayout.addLayout(self.opacityLayout)

        self.imageTreeView = QTreeView(SceneTab)
        self.imageTreeView.setObjectName(u"imageTreeView")
        self.imageTreeView.setSelectionMode(QAbstractItemView.SingleSelection)
        self.imageTreeView.setIndentation(5)

        self.verticalLayout.addWidget(self.imageTreeView)

        self.buttonsLayout = QHBoxLayout()
        self.buttonsLayout.setObjectName(u"buttonsLayout")
        self.addButton = QToolButton(SceneTab)
        self.addButton.setObjectName(u"addButton")
        icon = QIcon()
        icon.addFile(
            u":/icons/icons/baseline-add_circle-24px.svg",
            QSize(),
            QIcon.Normal,
            QIcon.Off,
        )
        self.addButton.setIcon(icon)
        self.addButton.setIconSize(QSize(24, 24))

        self.buttonsLayout.addWidget(self.addButton)

        self.upButton = QToolButton(SceneTab)
        self.upButton.setObjectName(u"upButton")
        self.upButton.setEnabled(True)
        icon1 = QIcon()
        icon1.addFile(
            u":/icons/icons/baseline-arrow_upward-24px.svg",
            QSize(),
            QIcon.Normal,
            QIcon.Off,
        )
        self.upButton.setIcon(icon1)
        self.upButton.setIconSize(QSize(24, 24))

        self.buttonsLayout.addWidget(self.upButton)

        self.downButton = QToolButton(SceneTab)
        self.downButton.setObjectName(u"downButton")
        icon2 = QIcon()
        icon2.addFile(
            u":/icons/icons/baseline-arrow_downward-24px.svg",
            QSize(),
            QIcon.Normal,
            QIcon.Off,
        )
        self.downButton.setIcon(icon2)
        self.downButton.setIconSize(QSize(24, 24))

        self.buttonsLayout.addWidget(self.downButton)

        self.deleteButton = QToolButton(SceneTab)
        self.deleteButton.setObjectName(u"deleteButton")
        icon3 = QIcon()
        icon3.addFile(
            u":/icons/icons/baseline-delete-24px.svg", QSize(), QIcon.Normal, QIcon.Off
        )
        self.deleteButton.setIcon(icon3)
        self.deleteButton.setIconSize(QSize(24, 24))

        self.buttonsLayout.addWidget(self.deleteButton)

        self.verticalLayout.addLayout(self.buttonsLayout)

        self.line = QFrame(SceneTab)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.HLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.verticalLayout.addWidget(self.line)

        self.toolboxWidget = QWidget(SceneTab)
        self.toolboxWidget.setObjectName(u"toolboxWidget")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(100)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.toolboxWidget.sizePolicy().hasHeightForWidth()
        )
        self.toolboxWidget.setSizePolicy(sizePolicy)
        self.verticalLayout_7 = QVBoxLayout(self.toolboxWidget)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(0, 0, 0, 0)
        self.toolsLabel = QLabel(self.toolboxWidget)
        self.toolsLabel.setObjectName(u"toolsLabel")
        self.toolsLabel.setMinimumSize(QSize(0, 0))
        font1 = QFont()
        font1.setPointSize(10)
        font1.setBold(True)
        self.toolsLabel.setFont(font1)
        self.toolsLabel.setAlignment(Qt.AlignLeading | Qt.AlignLeft | Qt.AlignVCenter)

        self.verticalLayout_7.addWidget(self.toolsLabel)

        self.toolsWidget = QWidget(self.toolboxWidget)
        self.toolsWidget.setObjectName(u"toolsWidget")
        sizePolicy1 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.toolsWidget.sizePolicy().hasHeightForWidth())
        self.toolsWidget.setSizePolicy(sizePolicy1)
        self.gridLayout = QGridLayout(self.toolsWidget)
        self.gridLayout.setObjectName(u"gridLayout")

        self.verticalLayout_7.addWidget(self.toolsWidget)

        self.line_2 = QFrame(self.toolboxWidget)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShape(QFrame.HLine)
        self.line_2.setFrameShadow(QFrame.Sunken)

        self.verticalLayout_7.addWidget(self.line_2)

        self.optionsLabel = QLabel(self.toolboxWidget)
        self.optionsLabel.setObjectName(u"optionsLabel")
        self.optionsLabel.setMinimumSize(QSize(0, 0))
        self.optionsLabel.setFont(font1)
        self.optionsLabel.setAlignment(Qt.AlignLeading | Qt.AlignLeft | Qt.AlignVCenter)

        self.verticalLayout_7.addWidget(self.optionsLabel)

        self.optionsWidget = QWidget(self.toolboxWidget)
        self.optionsWidget.setObjectName(u"optionsWidget")
        sizePolicy2 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.MinimumExpanding)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(1)
        sizePolicy2.setHeightForWidth(
            self.optionsWidget.sizePolicy().hasHeightForWidth()
        )
        self.optionsWidget.setSizePolicy(sizePolicy2)
        self.verticalLayout_2 = QVBoxLayout(self.optionsWidget)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")

        self.verticalLayout_7.addWidget(self.optionsWidget)

        self.verticalLayout.addWidget(self.toolboxWidget)

        self.verticalSpacer = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self.verticalLayout.addItem(self.verticalSpacer)

        self.retranslateUi(SceneTab)

        QMetaObject.connectSlotsByName(SceneTab)

    # setupUi

    def retranslateUi(self, SceneTab):
        SceneTab.setWindowTitle(QCoreApplication.translate("SceneTab", u"Form", None))
        self.opacitySliderLabel.setText(
            QCoreApplication.translate("SceneTab", u"Opacity", None)
        )
        # if QT_CONFIG(tooltip)
        self.opacitySlider.setToolTip(
            QCoreApplication.translate(
                "SceneTab", u"Change the opacity for all annotations", None
            )
        )
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(tooltip)
        self.addButton.setToolTip(
            QCoreApplication.translate("SceneTab", u"Add new Annotation Layers", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.addButton.setText(QCoreApplication.translate("SceneTab", u"...", None))
        # if QT_CONFIG(tooltip)
        self.upButton.setToolTip(
            QCoreApplication.translate("SceneTab", u"Move selected layer up", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.upButton.setText(QCoreApplication.translate("SceneTab", u"...", None))
        # if QT_CONFIG(tooltip)
        self.downButton.setToolTip(
            QCoreApplication.translate("SceneTab", u"Move selected layer down", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.downButton.setText(QCoreApplication.translate("SceneTab", u"...", None))
        # if QT_CONFIG(tooltip)
        self.deleteButton.setToolTip(
            QCoreApplication.translate("SceneTab", u"Delete selected layer", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.deleteButton.setText(QCoreApplication.translate("SceneTab", u"...", None))
        self.toolsLabel.setText(QCoreApplication.translate("SceneTab", u"Tools:", None))
        self.optionsLabel.setText(
            QCoreApplication.translate("SceneTab", u"Options:", None)
        )

    # retranslateUi



from .inspection import Inspection
from .pen import Pen
from .splinetool import Spline


def line_tools():
    return {"inspection": Inspection(), "spline": Spline()}


def area_tools():
    return {"inspection": Inspection(), "pen": Pen()}


def basic_tools():
    return {"inspection": Inspection()}



import logging
import os
import sys
from functools import partial
from pathlib import Path

import eyepy as ep
import requests
from packaging import version
from PySide6 import QtWidgets
from PySide6.QtCore import QCoreApplication, QSize
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QFileDialog, QMessageBox

import eyelab as el
from eyelab.config import EYELAB_FOLDER
from eyelab.dialogs.help import (
    AreaAnnotationHelp,
    IntroductionHelp,
    LayerAnnotationHelp,
    ShortcutHelp,
)
from eyelab.views.ui.ui_main_window import Ui_MainWindow
from eyelab.views.workspace import Workspace


class eyelab(QtWidgets.QMainWindow, Ui_MainWindow):
    """Create the main window that stores all of the widgets necessary for the application."""

    def __init__(self, parent=None):
        """Initialize the components of the main window."""
        super().__init__(parent)
        self.setupUi(self)

        self.save_path = None

        self.action_import_vol.triggered.connect(
            partial(self.import_data, method=ep.import_heyex_vol, format="file")
        )
        self.action_import_hexml.triggered.connect(
            partial(self.import_data, method=ep.import_heyex_xml, format="folder")
        )
        self.action_import_retouch.triggered.connect(
            partial(self.import_data, method=ep.import_retouch, format="folder")
        )
        self.action_import_duke.triggered.connect(
            partial(self.import_data, method=ep.import_duke_mat, format="file")
        )
        self.action_import_bsfolder.triggered.connect(
            partial(self.import_data, method=ep.import_bscan_folder, format="folder")
        )
        # self.action_save_annotations.triggered.connect(self.save_annotations)
        # self.action_save_annotations_as.triggered.connect(
        #    partial(self.save_annotations, save_as=True)
        # )
        # self.action_load_annotations.triggered.connect(self.load_annotations)
        self.action_open.triggered.connect(self.load)
        self.action_save.triggered.connect(self.save)
        self.action_save_as.triggered.connect(partial(self.save, save_as=True))
        self.action_shortcut_sheet.triggered.connect(
            lambda: self.open_help("shortcuts")
        )
        self.action_area_annotation_guide.triggered.connect(
            lambda: self.open_help("area_annotation")
        )
        self.action_layer_annotation_guide.triggered.connect(
            lambda: self.open_help("layer_annotation")
        )
        self.action_registration_guide.triggered.connect(
            lambda: self.open_help("registration")
        )
        self.action_introduction.triggered.connect(
            lambda: self.open_help("introduction")
        )

        self.workspace = Workspace(parent=self)
        self.workspace.undo_stack.cleanChanged.connect(self.toggle_save)

        self._edit_menu_setup()
        self.setCentralWidget(self.workspace)
        self.check_version()

        # hide options for loading/saving annotations only
        self.menuAnnotations.deleteLater()

        from eyepy.data import load

        ev = load("drusen_patient")
        ev.add_voxel_annotation(
            ep.drusen(ev.layers["RPE"], ev.layers["BM"], ev.shape), name="Drusen"
        )
        self.workspace.set_data(ev)
        self.statusBar().showMessage("Ready")

    def toggle_save(self, clean):
        if clean:
            self.action_save.setEnabled(False)
            self.action_save_as.setEnabled(False)
            self.setWindowTitle(self.windowTitle().rstrip("*"))
        else:
            self.action_save.setEnabled(True)
            self.action_save_as.setEnabled(True)
            self.setWindowTitle(self.windowTitle().rstrip("*") + "*")

    def _edit_menu_setup(self):
        self.action_undo = self.workspace.undo_stack.createUndoAction(self)
        self.action_redo = self.workspace.undo_stack.createRedoAction(self)

        undo_icon = QIcon()
        undo_icon.addFile(
            ":/icons/icons/baseline-undo-24px.svg", QSize(), QIcon.Normal, QIcon.Off
        )
        self.action_undo.setIcon(undo_icon)

        redo_icon = QIcon()
        redo_icon.addFile(
            ":/icons/icons/baseline-redo-24px.svg", QSize(), QIcon.Normal, QIcon.Off
        )
        self.action_redo.setIcon(redo_icon)

        self.action_undo.setShortcut(
            QCoreApplication.translate("MainWindow", "Ctrl+Z", None)
        )
        self.action_redo.setShortcut(
            QCoreApplication.translate("MainWindow", "Ctrl+Y", None)
        )

        self.menuEdit.addAction(self.action_undo)
        self.menuEdit.addAction(self.action_redo)

    def check_version(self):
        latest_url = "https://github.com/MedVisBonn/eyelab/releases/latest"
        current_version = f"v{el.__version__}"

        try:
            latest_version = (requests.get(latest_url).url).split("/")[-1]
        except requests.ConnectionError:
            return

        if version.parse(current_version) < version.parse(latest_version):
            msgBox = QMessageBox()
            msgBox.setText("A new version of EyeLab is available.")
            msgBox.setInformativeText(
                f"You are using version {current_version}. The latest version is {latest_version} and can be found <a href='{latest_url}'>here</a>."
            )
            msgBox.setStandardButtons(QMessageBox.Ok)
            msgBox.setDefaultButton(QMessageBox.Ok)
            ret = msgBox.exec()

    @property
    def start_dir(self):
        if self.save_path:
            return str(Path(self.save_path).parent)
        else:
            return str(Path.home())

    def import_data(self, method, format):
        if not self.workspace.data is None:
            message = "The import replaces all data you have in your workspace. Do you want to proceed?"
            ret = QMessageBox.question(
                self, "EyeLab", message, QMessageBox.Ok | QMessageBox.Cancel
            )

            if ret == QMessageBox.Cancel:
                return

        if format == "file":
            path = QFileDialog.getOpenFileName(dir=self.start_dir)[0]
        elif format == "folder":
            path = QFileDialog.getExistingDirectory(dir=self.start_dir)

        self.save_path = None
        self.workspace.set_data(method(path))

    def open_help(self, topic):
        if topic == "introduction":
            dialog = IntroductionHelp(self)
        elif topic == "shortcuts":
            dialog = ShortcutHelp(self)
        elif topic == "area_annotation":
            dialog = AreaAnnotationHelp(self)
        elif topic == "layer_annotation":
            dialog = LayerAnnotationHelp(self)
        # elif topic == "registration":
        #    dialog = RegistrationHelp(self)
        else:
            raise ValueError("topic not available")

        if dialog.exec() == QtWidgets.QDialog.Rejected:
            pass

    def get_save_location(self):
        if "PYCHARM_HOSTED" in os.environ:
            options = QFileDialog.DontUseNativeDialog
        else:
            options = QFileDialog.Options()
        save_path, filter = QFileDialog.getSaveFileName(
            parent=self, caption="Save", filter="Eye files (*.eye)", options=options
        )
        if save_path:
            if not save_path.endswith(".eye"):
                save_path = save_path + ".eye"
            self.save_path = save_path
        return self.save_path

    def save(self, save_as=False):
        if self.save_path is None or save_as is True:
            path = self.get_save_location()
            if path:
                self.workspace.data.save(self.save_path)

    def load(self):
        if self.workspace.data is not None:
            message = "Loading data replaces the current data. Do you want to proceed?"
            ret = QMessageBox.question(
                self, "EyeLab", message, QMessageBox.Ok | QMessageBox.Cancel
            )

            if ret == QMessageBox.Cancel:
                return

        path = QFileDialog.getOpenFileName(
            parent=self,
            caption="Load",
            dir=self.start_dir,
            filter="Eye files (*.eye)",
        )[0]
        if path == "":
            return
        if not path.startswith("/run/user"):
            self.save_path = path
        else:
            self.save_path = None
        ev = ep.EyeVolume.load(path)
        self.workspace.set_data(ev)

    def save_annotations(self, save_as=False):
        if self.save_path is None or save_as is True:
            path = self.get_save_location()
            if path == "":
                return

        self.workspace.data.save_annotations(self.save_path)

    def load_annotations(self):
        if self.workspace.data is None:
            message = (
                "Annotations can only be loaded after importing the corresponding data."
            )
            QMessageBox.information(self, "EyeLab", message, QMessageBox.Ok)
            return
        message = "Loading annotations replaces all current annotations. Do you want to proceed?"
        ret = QMessageBox.question(
            self, "EyeLab", message, QMessageBox.Ok | QMessageBox.Cancel
        )
        if ret == QMessageBox.Cancel:
            return
        path = QFileDialog.getOpenFileName()[0]
        self.workspace.data.load_annotations(path)


def main(log_level=logging.DEBUG):
    # create logger for "oat" application
    logger = logging.getLogger("eyelab")
    logger.setLevel(logging.DEBUG)

    # create file handler which logs debug messages
    fh = logging.FileHandler(EYELAB_FOLDER / "eyelab.log")
    fh.setLevel(logging.DEBUG)
    file_formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
    )
    fh.setFormatter(file_formatter)
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    cmd_formatter = logging.Formatter("%(levelname)s - %(name)s - %(message)s")
    ch.setFormatter(cmd_formatter)
    # add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)

    logger.info("Starting Application.")

    application = QtWidgets.QApplication(sys.argv)

    window = eyelab()
    # desktop = QtGui.QScreen().availableGeometry()
    # width = (desktop.width() - window.width()) / 2
    # height = (desktop.height() - window.height()) / 2
    # window.show()
    # window.move(width, height)
    window.showMaximized()

    sys.exit(application.exec())


if __name__ == "__main__":
    main()



from PySide6.QtGui import QUndoStack

_undo_stacks = {}


def get_undo_stack(name, parent=None):
    if name in _undo_stacks:
        return _undo_stacks[name]
    else:
        _undo_stacks[name] = QUndoStack(parent)
        return _undo_stacks[name]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_notification_diaolog.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore


class Ui_NotificationDialog(object):
    def setupUi(self, NotificationDialog):
        if not NotificationDialog.objectName():
            NotificationDialog.setObjectName(u"NotificationDialog")
        NotificationDialog.resize(297, 136)
        self.verticalLayout = QVBoxLayout(NotificationDialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.messageLabel = QLabel(NotificationDialog)
        self.messageLabel.setObjectName(u"messageLabel")
        self.messageLabel.setAlignment(Qt.AlignLeading | Qt.AlignLeft | Qt.AlignVCenter)
        self.messageLabel.setWordWrap(True)
        self.messageLabel.setMargin(5)

        self.verticalLayout.addWidget(self.messageLabel)

        self.buttonBox = QDialogButtonBox(NotificationDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(NotificationDialog)
        self.buttonBox.accepted.connect(NotificationDialog.accept)
        self.buttonBox.rejected.connect(NotificationDialog.reject)

        QMetaObject.connectSlotsByName(NotificationDialog)

    # setupUi

    def retranslateUi(self, NotificationDialog):
        NotificationDialog.setWindowTitle(
            QCoreApplication.translate("NotificationDialog", u"Notification", None)
        )
        self.messageLabel.setText("")

    # retranslateUi



import logging

import eyepy as ep
from PySide6 import QtCore, QtWidgets

from eyelab.views.graphicsview import CustomGraphicsView
from eyelab.views.ui.ui_data_view import Ui_DataView

logger = logging.getLogger("eyelab.workspace.dataview")


class DataView(QtWidgets.QWidget, Ui_DataView):
    cursorPosChanged = QtCore.Signal(QtCore.QPointF, CustomGraphicsView)

    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.setupUi(self)

        self.graphicsViewVolume.cursorPosChanged.connect(self.emit_volume_pos)
        self.graphicsViewLocalizer.cursorPosChanged.connect(self.emit_localizer_pos)

        self.graphicsViewLocalizer.hide()
        self.graphicsViewVolume.hide()

        self.data = None

    def set_data(self, data: ep.EyeVolume):
        logger.debug("DataView: set_data")
        self.data = data
        self.graphicsViewVolume.set_data(self.data, name="Volume")
        self.graphicsViewLocalizer.set_data(self.data.localizer, name="Enface")

        self.graphicsViewLocalizer.show()
        self.graphicsViewVolume.show()

    def emit_volume_pos(self, pos, sender):
        # The position coming from the volume is in the localizer space.
        # Since other views can only map from the localizer to their space
        # replace the sender here.
        self.cursorPosChanged.emit(pos, self.graphicsViewLocalizer)
        self.graphicsViewLocalizer.set_fake_cursor(pos)

    def emit_localizer_pos(self, pos, sender):
        self.graphicsViewVolume.set_fake_cursor(pos, sender)
        self.cursorPosChanged.emit(pos, sender)

    def set_fake_cursor(self, pos, sender):
        pos = self.graphicsViewLocalizer.map_from_sender(pos, sender)
        self.graphicsViewLocalizer.set_fake_cursor(pos, self.graphicsViewLocalizer)
        self.graphicsViewVolume.set_fake_cursor(pos, self.graphicsViewLocalizer)



from PySide6 import QtCore, QtGui, QtWidgets


class PaintPreview(QtWidgets.QGraphicsItem):
    def __init__(self, settings_widget, parent=None):
        super().__init__(parent)
        self.settings_widget = settings_widget
        self.init_preview()

    def init_preview(self):
        pass

    def boundingRect(self) -> QtCore.QRectF:
        return QtCore.QRectF(0.0, 0.0, 0.0, 0.0)

    def paint(
        self, painter: QtGui.QPainter, option: "QStyleOptionGraphicsItem", widget
    ) -> None:
        return None


class Inspection(object):
    def __init__(self):
        """ """
        self.name = "inspection"
        self.cursor = self.get_cursor()
        self.button = self.get_tool_button()
        self.hot_key = None
        self.options_widget = QtWidgets.QWidget()
        self.paint_preview = PaintPreview(self.options_widget)

    def get_tool_button(self):
        button = QtWidgets.QToolButton()
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(":/icons/icons/navigation.svg"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        button.setIcon(icon)
        button.setIconSize(QtCore.QSize(24, 24))
        button.setCheckable(True)
        button.setObjectName("inspectionButton")
        button.setToolTip("Inspection Tool")
        return button

    def enable(self):
        pass

    def disable(self):
        pass

    def get_cursor(self):
        return QtGui.QCursor(
            QtGui.QPixmap(":/cursors/cursors/navigation_cursor.svg"), hotX=0, hotY=0
        )

    @staticmethod
    def mouse_move_handler(gitem, event):
        pass

    @staticmethod
    def mouse_press_handler(gitem, event):
        pass

    @staticmethod
    def mouse_doubleclick_handler(gitem, event):
        pass

    @staticmethod
    def mouse_release_handler(gitem, event):
        pass

    @staticmethod
    def mouse_release_handler(gitem, event):
        pass

    @staticmethod
    def key_press_handler(gitem, event):
        pass

    @staticmethod
    def key_release_handler(gitem, event):
        pass



from PySide6 import QtCore, QtWidgets


class ItemGroup(QtWidgets.QGraphicsItem):
    def __init__(self, meta: dict, parent=None):
        """Provide data to create a new annotation or the id of an existing
        annotation.
        """
        super().__init__(parent=parent)

        self.meta = meta
        self.setFlag(QtWidgets.QGraphicsItem.ItemHasNoContents, True)

    def update(self):
        if self.meta:
            self.setVisible(self.meta["visible"])
            self.setZValue(self.meta["z_value"])

        for item in self.childItems():
            item.update()
        super().update

    def boundingRect(self) -> QtCore.QRectF:
        return self.childrenBoundingRect()



import json
import logging
from typing import List, Tuple

from PySide6 import QtCore, QtWidgets

from eyelab.views.ui.ui_add_annotation_dialog import Ui_AddAnnotationDialog

# from eyelab.models.treeview.itemmodel import TreeItemModel


logger = logging.getLogger(__name__)


class AddAnnotationDialog(QtWidgets.QDialog, Ui_AddAnnotationDialog):
    def __init__(self, model: "TreeItemModel", options: List[Tuple], parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.model = model
        self.options = options
        self.set_type_box()

        self.buttonBox.accepted.connect(self.add_annotation)
        self.buttonBox.rejected.connect(self.close)

    def set_type_box(self):
        for name, func in self.options.items():
            self.typeBox.addItem(name)

    def add_annotation(self):
        logger.debug("Add annotation layers")
        name = self.nameEdit.text()
        type = self.typeBox.currentText()

        self.options[type](name)



import logging

import numpy as np
from PySide6.QtCore import QLineF, QPointF, Qt
from PySide6.QtGui import QUndoCommand
from PySide6.QtWidgets import QGraphicsItem

from eyelab.models.treeview import layeritem

logger = logging.getLogger(__name__)


class AddLayeritem(QUndoCommand):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setText("Add Layer")

    def redo(self):
        logger.debug(f"Redo: {self.text()}")

    def undo(self):
        logger.debug(f"Undo: {self.text()}")


class DeleteLayeritem(QUndoCommand):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setText("Delete Layer")

    def redo(self):
        logger.debug(f"Redo: {self.text()}")

    def undo(self):
        logger.debug(f"Undo: {self.text()}")


class ChangePolygon(QUndoCommand):
    def __init__(
        self, polygon: "layeritem.PolygonPath", start=None, end=None, parent=None
    ):
        self.polygon = polygon
        self.layeritem = polygon.layer_item
        self.new_start = start if start else self.polygon.start
        self.new_end = end if end else self.polygon.end

        self.old_start = self.polygon.start
        self.old_end = self.polygon.end

        super().__init__(parent)
        self.setText("Change Layer Region")

    def redo(self):
        logger.debug(f"Redo: {self.text()}")
        self.polygon.start = self.new_start
        self.polygon.end = self.new_end

    def undo(self):
        logger.debug(f"Undo: {self.text()}")
        self.polygon.start = self.old_start
        self.polygon.end = self.old_end

    def id(self):
        return 3

    def mergeWith(self, other: QUndoCommand) -> bool:
        if self.polygon != other.polygon:
            return False

        self.new_start = other.new_start
        self.new_end = other.new_end
        return True


class AddPolygon(QUndoCommand):
    def __init__(self, layer_item: "layeritem.LayerItem", start, end, parent=None):
        self.layeritem = layer_item
        self.start = start
        self.end = end

        self.polygon = layeritem.PolygonPath(
            None, self.layeritem.height_map, self.start, self.end
        )

        super().__init__(parent)
        self.setText("Add Layer Region")

    def redo(self):
        logger.debug(f"Redo: {self.text()}")
        polygons = sorted(
            self.layeritem.polygons + [self.polygon], key=lambda x: x.start.x()
        )
        self.index = polygons.index(self.polygon)
        self.layeritem.polygons.insert(self.index, self.polygon)
        self.polygon.setParentItem(self.layeritem)

    def undo(self):
        logger.debug(f"Undo: {self.text()}")
        self.layeritem.polygons.pop(self.index)
        self.polygon.scene().removeItem(self.polygon)


class UpdateLayerArray(QUndoCommand):
    def __init__(self, bspline: "layeritem.CubicSpline", parent=None):
        self.bspline = bspline

        self.layeritem = None
        self.mapping = None
        self.old_mapping = None

        super().__init__(parent)
        self.setText("Update Layer Array")

    def redo(self):
        logger.debug(f"Redo: {self.text()}")
        self.layeritem = (
            self.bspline.layer_item if not self.layeritem else self.layeritem
        )
        self.mapping = self.bspline.indices() if not self.mapping else self.mapping
        self.old_mapping = (
            {x: self.layeritem.height_map[x] for x in self.mapping.keys()}
            if not self.old_mapping
            else self.old_mapping
        )
        for x in self.mapping:
            self.layeritem.height_map[x] = self.mapping[x]

    def undo(self):
        logger.debug(f"Undo: {self.text()}")
        for x in self.old_mapping:
            self.layeritem.height_map[x] = self.old_mapping[x]

    def id(self):
        return 4

    def mergeWith(self, other: QUndoCommand) -> bool:
        if self.bspline != other.bspline:
            return False

        self.mapping = {**self.mapping, **other.mapping}
        self.old_mapping = {**other.old_mapping, **self.old_mapping}
        return True


class MoveKnot(QUndoCommand):
    def __init__(self, knot: QGraphicsItem, new_pos: QPointF, optimize_neighbours=True):
        self.knot = knot
        self.bspline = self.knot.bspline
        self.index = self.bspline.knots.index(self.knot)
        self.optimize_neighbours = optimize_neighbours

        # Make sure knots ara x centered in pixel
        new_pos.setX(np.floor(new_pos.x()) + 0.5)
        self.new_pos = new_pos
        self.old_pos = self.knot.center
        super().__init__()
        self.setText("Move Knot")

        # Create child command for optimizing control points
        self.optimize_cps = OptimizeControlPoints(
            self.knot, self.bspline, propagate=self.optimize_neighbours, parent=self
        )

        self.change_p1, self.change_p2 = None, None
        self.deleted_polygon = None
        (
            left_neighbours,
            right_neighbours,
        ) = self.bspline.layer_item.get_neighbour_polygons(self.bspline)

        if self.knot is self.bspline.knots[-1] and right_neighbours:
            for n in right_neighbours:
                # Find the closest neighbour that is not covered after the move
                if n.end.x() >= self.new_pos.x():
                    # Change the neighbour if the new pos is in its region
                    if self.old_pos.x() == n.start.x() or self.new_pos in n:
                        self.change_p1 = ChangePolygon(
                            n, start=self.new_pos, parent=self
                        )
                    break
                else:
                    # Delete polygon if it is covered
                    self.deleted_polygon = DeletePolygon(n, parent=self)

        if self.knot is self.bspline.knots[0] and left_neighbours:
            for n in left_neighbours:
                # Find the closest neighbour that is not covered after the move
                if n.start.x() <= self.new_pos.x():
                    if self.old_pos.x() == n.end.x() or self.new_pos in n:
                        self.change_p2 = ChangePolygon(n, end=self.new_pos, parent=self)
                    break
                else:
                    self.deleted_polygon = DeletePolygon(n, parent=self)

        self.update_array = UpdateLayerArray(self.bspline, parent=self)

        self.added_polygon = None
        if (
            self.knot is self.bspline.knots[-1]
            and self.new_pos.x() < self.old_pos.x()
            and not self.change_p1
        ):
            self.added_polygon = AddPolygon(
                self.bspline.layer_item,
                start=self.new_pos,
                end=self.old_pos,
                parent=self,
            )

        if (
            self.knot is self.bspline.knots[0]
            and self.new_pos.x() > self.old_pos.x()
            and not self.change_p2
        ):
            self.added_polygon = AddPolygon(
                self.bspline.layer_item,
                start=self.old_pos,
                end=self.new_pos,
                parent=self,
            )

    def redo(self) -> None:
        logger.debug(f"Redo: {self.text()}")
        self.knot.setPos(self.new_pos)
        self.knot.sync()
        super().redo()
        self.knot.bspline.update()

        if self.change_p1:
            self.change_p1.polygon.update()
        if self.change_p2:
            self.change_p2.polygon.update()
        if self.added_polygon:
            self.bspline.layer_item.update()

    def undo(self) -> None:
        logger.debug(f"Undo: {self.text()}")
        self.knot.setPos(self.old_pos)
        self.knot.sync()
        super().undo()
        self.knot.bspline.update()

    def mergeWith(self, other: "MoveKnot") -> bool:
        p1s = [c.polygon if c else None for c in [self.change_p1, other.change_p1]]
        p2s = [c.polygon if c else None for c in [self.change_p2, other.change_p2]]
        # Do only merge if the same knot is moved
        if other.knot != self.knot:
            return False
        # Do only merge if the new command does not delete a polygon
        if not other.deleted_polygon is None:
            return False
        # Do only merge if the old command did not add a polygon
        if not (self.added_polygon is None and other.added_polygon is None):
            return False
        # Do only merge if neighbouring polygons are the same or one of them is None
        if not (p1s[0] is None or p1s[1] is None or p1s[0] is p1s[1]):
            return False
        if not (p2s[0] is None or p2s[1] is None or p2s[0] is p2s[1]):
            return False

        self.new_pos = other.new_pos

        self.optimize_cps.mergeWith(other.optimize_cps)
        if self.change_p1 and other.change_p1:
            self.change_p1.mergeWith(other.change_p1)
        if self.change_p2 and other.change_p2:
            self.change_p2.mergeWith(other.change_p2)

        self.update_array.mergeWith(other.update_array)
        return True

    def id(self):
        return 1


class MoveControlKnot(QUndoCommand):
    def __init__(self, item: QGraphicsItem, new_pos: QPointF):
        self.cp = item
        self.knot = self.cp.parentItem()
        cp_in = self.knot.cp_in
        cp_out = self.knot.cp_out
        self.other_cp = cp_in if not self.cp is cp_in else cp_out

        self.new_pos = new_pos
        self.old_pos = self.cp.pos()

        # Make sure control points move together to keep the curve smooth
        line = QLineF(self.cp.center, self.knot.center)
        line2 = QLineF(self.knot.center, self.other_cp.center)
        line.setLength(line.length() + line2.length())

        self.other_old = self.other_cp.pos()
        self.other_new = self.knot.mapFromScene(line.p2())

        UpdateLayerArray(self.knot.bspline, parent=self)

        super().__init__()
        self.setText("Move Control Knot")

    def redo(self) -> None:
        logger.debug(f"Redo: {self.text()}")
        self.cp.setPos(self.new_pos)
        self.other_cp.setPos(self.other_new)

        super().redo()
        self.knot.bspline.update()
        self.knot.set_lines()

    def undo(self) -> None:
        logger.debug(f"Undo: {self.text()}")
        self.cp.setPos(self.old_pos)
        self.cp.setPos(self.old_pos)
        self.other_cp.setPos(self.other_old)

        super().undo()
        self.knot.bspline.update()
        self.knot.set_lines()

    def mergeWith(self, other: "MoveKnot") -> bool:
        # Make sure other is a MoveKnot command for the same QGraphicsItem
        if other.cp != self.cp:
            return False

        other.old_pos = self.old_pos
        other.other_old = self.other_old
        return True

    def id(self):
        return 2


class AddKnot(QUndoCommand):
    def __init__(
        self,
        bspline: "layeritem.CubicSpline",
        pos: QPointF,
        optimize_neighbours=True,
        parent=None,
    ):
        self.bspline = bspline
        self.layeritem = self.bspline.layer_item

        pos.setX(np.floor(pos.x()) + 0.5)
        self.pos = pos
        self.optimize_neighbours = optimize_neighbours
        super().__init__(parent)
        self.setText("Add Knot")

        # Create new knot
        knot_dict = {
            "knot_pos": (pos.x(), pos.y()),
            "cp_in_pos": (pos.x() - 10, pos.y()),
            "cp_out_pos": (pos.x() + 10, pos.y()),
        }

        self.new_knot = layeritem.CubicSplineKnotItem(parent=None, knot_dict=knot_dict)

        # Get insertion index ToDo: Replace with bisect insort when Pysid6 works on Python3.10
        i = 0
        for k in self.bspline._knots:
            if knot_dict["knot_pos"][0] > k["knot_pos"][0]:
                i += 1
                continue
            break
        self.index = i
        # Create child command for optimizing control points
        OptimizeControlPoints(
            self.new_knot, self.bspline, propagate=self.optimize_neighbours, parent=self
        )

        # Create child commands to change neighbouring polygons if necessary
        left_neighbours, right_neighbours = self.layeritem.get_neighbour_polygons(
            self.bspline
        )
        if self.index == len(self.bspline.knots) and right_neighbours:
            for n in right_neighbours:
                if self.pos in n:
                    ChangePolygon(n, start=self.pos, parent=self)
                    break
                elif self.pos.x() < n.start.x():
                    break
                else:
                    DeletePolygon(n, parent=self)

        if self.index == 0 and left_neighbours:
            for n in left_neighbours:
                if self.pos in n:
                    ChangePolygon(n, end=self.pos, parent=self)
                    break
                elif self.pos.x() > n.end.x():
                    break
                else:
                    DeletePolygon(n, parent=self)

        self.update_array = UpdateLayerArray(self.bspline, parent=self)

    def redo(self) -> None:
        logger.debug(f"Redo: {self.text()}")
        self.new_knot.setParentItem(self.bspline)
        self.bspline._knots.insert(self.index, self.new_knot.knot_dict)
        self.bspline.knots.insert(self.index, self.new_knot)

        if not self.bspline.control_points_visible:
            self.new_knot.hide_control_points()

        self.new_knot.setFocus(Qt.MouseFocusReason)
        super().redo()
        self.bspline.update()

    def undo(self) -> None:
        logger.debug(f"Undo: {self.text()}")
        super().undo()

        self.new_knot.scene().removeItem(self.new_knot)
        self.bspline.knots.pop(self.index)
        self.bspline._knots.pop(self.index)
        self.bspline.update()


class DeleteKnot(QUndoCommand):
    def __init__(self, knot: "layeritem.CubicSplineKnotItem", optimize_neighbours=True):
        self.knot = knot
        self.optimize_neighbours = optimize_neighbours
        super().__init__()
        self.setText("Delete Knot")

        self.bspline = self.knot.bspline
        self.layeritem = self.bspline.layer_item
        self.index = self.bspline.knots.index(self.knot)

        left = self.bspline.knots[self.index - 1] if self.index - 1 > 0 else None
        right = (
            self.bspline.knots[self.index + 1]
            if self.index + 1 < len(self.bspline.knots)
            else None
        )
        if left and self.optimize_neighbours:
            OptimizeControlPoints(left, self.bspline, propagate=False, parent=self)
        if right and self.optimize_neighbours:
            OptimizeControlPoints(right, self.bspline, propagate=False, parent=self)

        neighbours = self.layeritem.get_neighbour_elements(self.bspline)
        if len(self.bspline.knots) == 1:
            # Delete Bspline object when last knot is removed
            DeleteCurve(self.bspline, self)
        elif self.knot is self.bspline.knots[0]:
            p = neighbours[0] if neighbours[0].end == self.knot.center else None
            if p:
                ChangePolygon(p, end=self.bspline.knots[1].center, parent=self)
        elif self.knot is self.bspline.knots[-1]:
            p = neighbours[1] if neighbours[1].start == self.knot.center else None
            if p:
                ChangePolygon(p, start=self.bspline.knots[-2].center, parent=self)

    def redo(self) -> None:
        logger.debug(f"Redo: {self.text()}")
        if len(self.bspline.knots) > 1:
            self.bspline.scene().removeItem(self.knot)
            self.bspline.knots.pop(self.index)
            self.bspline._knots.pop(self.index)

            self.bspline.update()
        super().redo()

    def undo(self) -> None:
        logger.debug(f"Undo: {self.text()}")
        super().undo()
        # If Curve was not removed
        if self.bspline.parentItem():
            self.bspline.knots.insert(self.index, self.knot)
            self.bspline._knots.insert(self.index, self.knot.knot_dict)
            self.knot.setParentItem(self.bspline)

            self.bspline.update()


class DeleteCurve(QUndoCommand):
    def __init__(self, bspline: "layeritem.CubicSpline", parent=None):
        self.bspline = bspline
        self.layeritem = bspline.layer_item
        self.index = self.layeritem.cubic_splines.index(self.bspline)

        super().__init__(parent)
        self.setText("Delete Knot")

        # Rejoin Polygons
        p1 = [
            p for p in self.layeritem.polygons if p.end == self.bspline.knots[0].center
        ]
        p2 = [
            p
            for p in self.layeritem.polygons
            if p.start == self.bspline.knots[-1].center
        ]
        if p1 and p2:
            JoinPolygons(p1[0], p2[0], self)

    def redo(self) -> None:
        logger.debug(f"Redo: {self.text()}")
        self.layeritem.cubic_splines.pop(self.index)
        self.knots = self.layeritem.knots.pop(self.index)
        self.layeritem.scene().removeItem(self.bspline)

        super().redo()

    def undo(self) -> None:
        logger.debug(f"Undo: {self.text()}")
        super().undo()
        self.layeritem.cubic_splines.insert(self.index, self.bspline)
        self.layeritem.knots.insert(self.index, self.knots)
        self.bspline.setParentItem(self.layeritem)


class DeletePolygon(QUndoCommand):
    def __init__(self, polygon: "layeritem.PolygonPath", parent=None):
        self.polygon = polygon
        self.layeritem = self.polygon.layer_item

        super().__init__(parent)
        self.setText("Delete Layer Region")

        start, stop = self.polygon.x_region
        self.slice = np.s_[int(np.floor(start)) : int(np.ceil(stop))]

    def redo(self) -> None:
        logger.debug(f"Redo: {self.text()}")
        self.index = self.layeritem.polygons.index(self.polygon)
        self.layeritem.polygons.pop(self.index)
        self.heights = np.copy(self.polygon.heights[self.slice])
        self.polygon.heights[self.slice] = np.nan
        self.layeritem.scene().removeItem(self.polygon)

        super().redo()

    def undo(self) -> None:
        logger.debug(f"Undo: {self.text()}")
        super().undo()
        self.layeritem.polygons.insert(self.index, self.polygon)
        self.polygon.heights[self.slice] = self.heights
        self.polygon.setParentItem(self.layeritem)


class OptimizeControlPoints(QUndoCommand):
    def __init__(
        self,
        knot: "layeritem.CubicSplineKnotItem",
        bspline: "layeritem.CubicSpline",
        propagate=False,
        parent=None,
    ):
        self.knot = knot
        self.bspline = bspline

        self.propagate = propagate
        super().__init__(parent)
        self.setText("Optimize Controllpoints")

        self.in_pos, self.out_pos = None, None
        self.old_in_pos, self.old_out_pos = None, None
        self.left, self.right = None, None
        if self.propagate:
            if not self.knot in self.bspline.knots:
                knots = sorted(
                    self.bspline.knots + [self.knot],
                    key=lambda x: x.knot_dict["knot_pos"][0],
                )
            else:
                knots = self.bspline.knots

            index = knots.index(self.knot)
            left = knots[index - 1] if index > 0 else None
            right = knots[index + 1] if index < len(knots) - 1 else None

            if left:
                self.left = OptimizeControlPoints(
                    left, self.bspline, propagate=False, parent=self
                )
            if right:
                self.right = OptimizeControlPoints(
                    right, self.bspline, propagate=False, parent=self
                )

    def redo(self) -> None:
        logger.debug(f"Redo: {self.text()}")
        if not self.old_in_pos:
            self.old_in_pos, self.old_out_pos = (
                self.knot.cp_in_pos,
                self.knot.cp_out_pos,
            )
        if not self.in_pos:
            self.in_pos, self.out_pos = self.bspline.optimize_controllpoints(self.knot)

        self.knot.cp_in = self.in_pos
        self.knot.cp_out = self.out_pos
        super().redo()

    def undo(self) -> None:
        logger.debug(f"Undo: {self.text()}")
        self.knot.cp_in = self.old_in_pos
        self.knot.cp_out = self.old_out_pos
        super().undo()

    def id(self) -> int:
        return 5

    def mergeWith(self, other: QUndoCommand) -> bool:
        if self.knot != other.knot:
            return False
        if not self.left and other.left:
            return False
        if not self.right and other.right:
            return False

        self.in_pos = other.in_pos
        self.out_pos = other.out_pos

        if self.propagate:
            if self.left:
                self.left.mergeWith(other.left)
            if self.right:
                self.right.mergeWith(other.right)
        return True


class AddCurve(QUndoCommand):
    def __init__(self, layer_item: "layeritem.LayerItem", pos: QPointF):
        self.layeritem = layer_item

        pos.setX(np.floor(pos.x()) + 0.5)
        self.pos = pos
        super().__init__()
        self.setText("New Curve")

        knot_dict = {
            "knot_pos": (pos.x(), pos.y()),
            "cp_in_pos": (pos.x() - 10, pos.y()),
            "cp_out_pos": (pos.x() + 10, pos.y()),
        }

        self.knots = []
        self.bspline = layeritem.CubicSpline(self.knots, None)
        self.new_knot = layeritem.CubicSplineKnotItem(parent=None, knot_dict=knot_dict)

        self.new_knot.setParentItem(self.bspline)
        self.bspline._knots.append(self.new_knot.knot_dict)
        self.bspline.knots.append(self.new_knot)

        if not self.bspline.control_points_visible:
            self.new_knot.hide_control_points()

        # Get insertion index
        i = 0
        for cs in self.layeritem.cubic_splines:
            if self.new_knot.center.x() > cs.start.x():
                i += 1
            else:
                break
        self.index = i

        # Split polygon if new Curve lies in it.
        polygon = [p for p in self.layeritem.polygons if self.pos in p]
        if polygon:
            polygon = polygon[0]
            SplitPolygons(polygon, self.pos, self)

        self.update_array = UpdateLayerArray(self.bspline, parent=self)

    def redo(self) -> None:
        logger.debug(f"Redo: {self.text()}")
        self.bspline.setParentItem(self.layeritem)
        self.layeritem.knots.insert(self.index, self.knots)
        self.layeritem.cubic_splines.insert(self.index, self.bspline)

        self.new_knot.setFocus(Qt.MouseFocusReason)

        super().redo()
        self.layeritem.update()

    def undo(self) -> None:
        logger.debug(f"Undo: {self.text()}")
        super().undo()
        self.bspline.scene().removeItem(self.bspline)
        self.layeritem.cubic_splines.pop(self.index)
        self.layeritem.knots.pop(self.index)
        self.layeritem.update()


class SplitPolygons(QUndoCommand):
    def __init__(self, polygon: "layeritem.PolygonPath", pos: QPointF, parent=None):
        self.polygon = polygon
        self.pos = pos

        self.layeritem = self.polygon.layer_item
        self.new_polygons = [
            layeritem.PolygonPath(
                None, self.polygon.heights, self.polygon.start, self.pos
            ),
            layeritem.PolygonPath(
                None, self.polygon.heights, self.pos, self.polygon.end
            ),
        ]
        self.index = self.layeritem.polygons.index(self.polygon)
        super().__init__(parent)
        self.setText("Split Polygons")

    def redo(self) -> None:
        logger.debug(f"Redo: {self.text()}")
        for p in self.new_polygons:
            p.setParentItem(self.polygon.parentItem())

        self.polygon.scene().removeItem(self.polygon)
        self.layeritem.polygons = (
            self.layeritem.polygons[: self.index]
            + self.new_polygons
            + self.layeritem.polygons[self.index + 1 :]
        )

    def undo(self) -> None:
        logger.debug(f"Undo: {self.text()}")
        self.polygon.setParentItem(self.new_polygons[0].parentItem())
        for p in self.new_polygons:
            p.scene().removeItem(p)

        self.layeritem.polygons = (
            self.layeritem.polygons[: self.index]
            + [self.polygon]
            + self.layeritem.polygons[self.index + 2 :]
        )


class JoinPolygons(QUndoCommand):
    def __init__(
        self, p1: "layeritem.PolygonPath", p2: "layeritem.PolygonPath", parent=None
    ):
        self.polygons = [p1, p2]

        self.layeritem = p1.layer_item
        self.new_polygon = layeritem.PolygonPath(None, p1.heights, p1.start, p2.end)
        self.index = self.layeritem.polygons.index(p1)

        super().__init__(parent)
        self.setText("Join Polygons")

    def redo(self) -> None:
        logger.debug(f"Redo: {self.text()}")
        self.new_polygon.setParentItem(self.polygons[0].parentItem())
        for p in self.polygons:
            p.scene().removeItem(p)

        self.layeritem.polygons = (
            self.layeritem.polygons[: self.index]
            + [self.new_polygon]
            + self.layeritem.polygons[self.index + 2 :]
        )
        self.layeritem.update()

    def undo(self) -> None:
        logger.debug(f"Undo: {self.text()}")
        for p in self.polygons:
            p.setParentItem(self.new_polygon.parentItem())

        self.new_polygon.scene().removeItem(self.new_polygon)
        self.layeritem.polygons = (
            self.layeritem.polygons[: self.index]
            + self.polygons
            + self.layeritem.polygons[self.index + 1 :]
        )
        self.layeritem.update()



from eyelab.dialogs.help.help_dialog import HelpWindow

help_text = """
# Area Annotation Guide

An area annotation in Eyelab is a pixel_map on top of the underlying image. Its color can
be changed, as well as its visibility.

## Adding a new Area Annotation

Create a new area using the (+) symbol in the overview on the right. Specify a name
and select Areas from the Type dropdown.

## Area Annotation

First select the area you want to annotate from the overview on the right. Selecting an
area automatically enables the **Pen** tool. You can now make annotations. The size of the
**Pen** tool can be changed on the right.

If you want to pan around in the image select the **Inspection** tool on the right.
"""


class AreaAnnotationHelp(HelpWindow):
    def __init__(self, parent):
        super().__init__(help_text, parent)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_data_view.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore

from eyelab.views.enface_view import EnfaceView
from eyelab.views.volume_view import VolumeView

from . import resources_rc


class Ui_DataView(object):
    def setupUi(self, DataView):
        if not DataView.objectName():
            DataView.setObjectName(u"DataView")
        DataView.resize(670, 200)
        sizePolicy = QSizePolicy(
            QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding
        )
        sizePolicy.setHorizontalStretch(5)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(DataView.sizePolicy().hasHeightForWidth())
        DataView.setSizePolicy(sizePolicy)
        self.horizontalLayout = QHBoxLayout(DataView)
        self.horizontalLayout.setSpacing(2)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.graphicsViewLocalizer = EnfaceView(DataView)
        self.graphicsViewLocalizer.setObjectName(u"graphicsViewLocalizer")
        sizePolicy1 = QSizePolicy(
            QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding
        )
        sizePolicy1.setHorizontalStretch(2)
        sizePolicy1.setVerticalStretch(2)
        sizePolicy1.setHeightForWidth(
            self.graphicsViewLocalizer.sizePolicy().hasHeightForWidth()
        )
        self.graphicsViewLocalizer.setSizePolicy(sizePolicy1)
        self.graphicsViewLocalizer.setMinimumSize(QSize(200, 200))

        self.horizontalLayout.addWidget(self.graphicsViewLocalizer)

        self.horizontalSpacer_3 = QSpacerItem(
            2, 20, QSizePolicy.Preferred, QSizePolicy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer_3)

        self.graphicsViewVolume = VolumeView(DataView)
        self.graphicsViewVolume.setObjectName(u"graphicsViewVolume")
        sizePolicy2 = QSizePolicy(
            QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding
        )
        sizePolicy2.setHorizontalStretch(4)
        sizePolicy2.setVerticalStretch(2)
        sizePolicy2.setHeightForWidth(
            self.graphicsViewVolume.sizePolicy().hasHeightForWidth()
        )
        self.graphicsViewVolume.setSizePolicy(sizePolicy2)
        self.graphicsViewVolume.setMinimumSize(QSize(400, 200))

        self.horizontalLayout.addWidget(self.graphicsViewVolume)

        self.retranslateUi(DataView)

        QMetaObject.connectSlotsByName(DataView)

    # setupUi

    def retranslateUi(self, DataView):
        DataView.setWindowTitle(QCoreApplication.translate("DataView", u"Form", None))

    # retranslateUi



from typing import Union

from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtWidgets import QWidget

from eyelab.models.layereditor import LayerEntry, LayerGroupEntry


class TreeItemDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent):
        super().__init__(parent)

    def paint(
        self,
        painter: QtGui.QPainter,
        option: "QStyleOptionViewItem",
        index: QtCore.QModelIndex,
    ) -> None:
        super().paint(painter, option, index)
        self.parent().openPersistentEditor(index)

    def createEditor(
        self, parent: QWidget, option: "QStyleOptionViewItem", index: QtCore.QModelIndex
    ) -> QWidget:
        # SceneTab: parent.parent().parent()
        if index.parent().parent() == QtCore.QModelIndex():
            self.editor = LayerGroupEntry(parent)
        else:
            self.editor = LayerEntry(parent)
        self.editor.editorChanged.connect(self.update_model)
        return self.editor

    @QtCore.Slot()
    def update_model(self):
        editor = self.sender()
        self.commitData.emit(editor)

    def setEditorData(self, editor: QWidget, index: QtCore.QModelIndex) -> None:
        data = index.model().data(index, QtCore.Qt.EditRole)
        editor.label.setText(str(data["name"]))
        editor.set_visible(data["visible"])
        if type(editor) == LayerEntry:
            editor.set_color(data["current_color"].upper())

    def setModelData(
        self,
        editor: Union[LayerEntry, LayerGroupEntry],
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ) -> None:
        data = {"visible": editor.visible, "name": editor.label.text()}
        if type(editor) is LayerEntry:
            data["current_color"] = editor.color

        model.setData(index, data, QtCore.Qt.EditRole)

    def sizeHint(
        self, option: "QStyleOptionViewItem", index: QtCore.QModelIndex
    ) -> QtCore.QSize:
        return LayerEntry(None).sizeHint()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_volumelocalizer_view.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from oat.modules.annotation.views.enface_view import EnfaceView
from oat.modules.annotation.views.volume_view import VolumeView
from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore

from . import resources_rc


class Ui_VolumeLocalizerView(object):
    def setupUi(self, VolumeLocalizerView):
        if not VolumeLocalizerView.objectName():
            VolumeLocalizerView.setObjectName(u"VolumeLocalizerView")
        VolumeLocalizerView.resize(670, 200)
        sizePolicy = QSizePolicy(
            QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding
        )
        sizePolicy.setHorizontalStretch(5)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(
            VolumeLocalizerView.sizePolicy().hasHeightForWidth()
        )
        VolumeLocalizerView.setSizePolicy(sizePolicy)
        self.horizontalLayout = QHBoxLayout(VolumeLocalizerView)
        self.horizontalLayout.setSpacing(2)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.graphicsViewLocalizer = EnfaceView(VolumeLocalizerView)
        self.graphicsViewLocalizer.setObjectName(u"graphicsViewLocalizer")
        sizePolicy1 = QSizePolicy(
            QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding
        )
        sizePolicy1.setHorizontalStretch(2)
        sizePolicy1.setVerticalStretch(2)
        sizePolicy1.setHeightForWidth(
            self.graphicsViewLocalizer.sizePolicy().hasHeightForWidth()
        )
        self.graphicsViewLocalizer.setSizePolicy(sizePolicy1)
        self.graphicsViewLocalizer.setMinimumSize(QSize(200, 200))

        self.horizontalLayout.addWidget(self.graphicsViewLocalizer)

        self.horizontalSpacer_3 = QSpacerItem(
            2, 20, QSizePolicy.Preferred, QSizePolicy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer_3)

        self.graphicsViewVolume = VolumeView(VolumeLocalizerView)
        self.graphicsViewVolume.setObjectName(u"graphicsViewVolume")
        sizePolicy2 = QSizePolicy(
            QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding
        )
        sizePolicy2.setHorizontalStretch(4)
        sizePolicy2.setVerticalStretch(2)
        sizePolicy2.setHeightForWidth(
            self.graphicsViewVolume.sizePolicy().hasHeightForWidth()
        )
        self.graphicsViewVolume.setSizePolicy(sizePolicy2)
        self.graphicsViewVolume.setMinimumSize(QSize(400, 200))

        self.horizontalLayout.addWidget(self.graphicsViewVolume)

        self.retranslateUi(VolumeLocalizerView)

        QMetaObject.connectSlotsByName(VolumeLocalizerView)

    # setupUi

    def retranslateUi(self, VolumeLocalizerView):
        VolumeLocalizerView.setWindowTitle(
            QCoreApplication.translate("VolumeLocalizerView", u"Form", None)
        )

    # retranslateUi



import collections
from typing import Union

import eyepy as ep
import numpy as np
from eyepy.core.utils import DynamicDefaultDict
from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import QAbstractItemModel, Qt

from eyelab.dialogs import AddAnnotationDialog
from eyelab.models.scene import CustomGraphicsScene
from eyelab.models.treeview.areaitem import AreaItem
from eyelab.models.treeview.itemgroup import ItemGroup
from eyelab.models.treeview.layeritem import LayerItem


class TreeItemGroup:
    def __init__(self, meta: dict, parent=None):
        self.itemData = meta
        self.parent = parent
        self.childItems = []

    def childNumber(self):
        if self.parent:
            return self.parent.childItems.index(self)
        return 0

    def child(self, number: int):
        if number < 0 or number > self.childCount() - 1:
            return None
        return self.childItems[number]

    def childCount(self):
        return len(self.childItems)

    def columnCount(self):
        return 1

    def data(self, column: str):
        if column in self.itemData:
            return self.itemData[column]
        raise Exception(f"column {column} not in data")

    def setData(self, column: str, value):
        if column in self.itemData:
            self.itemData[column] = value
            return True
        return False

    def appendChild(self, new_item: Union["TreeItem", "TreeItemGroup"]):
        items = self.childItems

        if items:
            z_value = float(items[-1].data("z_value") + 1)
        else:
            z_value = 0.0

        new_item.z_value = z_value

        new_item.parent = self
        self.childItems.append(new_item)

    def removeChildren(self, row: int, count: int):
        for i in range(row, row + count):
            self.childItems.pop(i)

    def switchChildren(self, row1: int, row2: int):
        child1 = self.child(row1)
        child2 = self.child(row2)

        child1_z = child1.data("z_value")
        child2_z = child2.data("z_value")
        child1.setData("z_value", child2_z)
        child2.setData("z_value", child1_z)


class TreeItem(TreeItemGroup):
    def __init__(self, data, parent=None):
        self.itemData = data.meta
        self.annotation = data
        self.parent = parent

    def childNumber(self):
        if self.parent:
            return self.parent.childItems.index(self)
        return 0

    def child(self, number: int):
        return None

    def childCount(self):
        return 0

    def data(self, column: str):
        if column in self.itemData:
            return self.itemData[column]
        raise Exception(f"column {column} not in data")

    def setData(self, column: str, value):
        if column == "name":
            self.itemData.name = value
            return True
        elif column in self.itemData:
            self.itemData[column] = value
            return True
        return False

    def appendChild(self, new_item):
        raise ValueError("You can only append childs to the TreeItemGroup")


class TreeItemModel(QtCore.QAbstractItemModel):
    def __init__(self, parent):
        super().__init__(parent=parent)

    def toggle_annotations(self):
        for row in range(self.tree_root.childCount()):
            index = self.index(row, 0)
            self.setData(index=index, data={"visible": not self.data(index)["visible"]})

    def rowCount(self, parent=QtCore.QModelIndex(), *args, **kwargs):
        parent_item = self.getItem(parent)
        return parent_item.childCount()

    def columnCount(self, parent=QtCore.QModelIndex(), *args, **kwargs):
        return self.tree_root.columnCount()

    def data(self, index: QtCore.QModelIndex(), role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.EditRole:
            item = self.getItem(index)
            return item.itemData
        if role == QtCore.Qt.DisplayRole:
            pass

    def index(self, row, column, parent=QtCore.QModelIndex(), *args, **kwargs):
        if not parent.isValid():
            parent_item = self.tree_root
        else:
            parent_item = parent.internalPointer()

        child_item = parent_item.child(row)
        if child_item:
            return self.createIndex(row, column, child_item)
        return QtCore.QModelIndex()

    def parent(self, index: QtCore.QModelIndex):
        if not index.isValid():
            return QtCore.QModelIndex()

        childItem = self.getItem(index)
        parentItem = childItem.parent

        if parentItem is None:
            return QtCore.QModelIndex()

        return self.createIndex(parentItem.childNumber(), 0, parentItem)

    def headerData(self, column, Qt_Orientation, role=None):
        return None

    def getItem(self, index: QtCore.QModelIndex):
        if index.isValid():
            item = index.internalPointer()
            if item:
                return item
        return self.tree_root

    def flags(self, index: QtCore.QModelIndex()):
        if not index.isValid():
            return QtCore.Qt.NoItemFlags

        return QtCore.Qt.ItemIsEditable | QAbstractItemModel.flags(self, index)

    # Provide support for editing and resizing
    def setData(self, index: QtCore.QModelIndex, data, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.EditRole:
            item = self.getItem(index)
            for k in data:
                item.itemData[k] = data[k]
            self.dataChanged.emit(index, index)
            return True
        return False

    def appendRow(self, data, parent=QtCore.QModelIndex()):
        self.beginInsertRows(parent, self.rowCount(parent), self.rowCount(parent))
        self.getItem(parent).appendChild(data)
        self.endInsertRows()

    def switchRows(self, row1, row2, parent=QtCore.QModelIndex()):
        self.beginMoveRows(parent, row1, row1, parent, row2 + 1)
        self.getItem(parent).switchChildren(row1, row2)
        self.endMoveRows()

    def removeRows(self, row, count, parent=QtCore.QModelIndex()):
        if parent.isValid() and parent.internalPointer() != self.tree_root:
            self.beginRemoveRows(parent, row, row + count - 1)
            parent = self.getItem(parent)
            parent.removeChildren(row, count)
            self.endRemoveRows()
            return True
        return False

    def _items_from_index(self, index):
        raise NotImplementedError

    def activate(self, index):
        scene_items = self._items_from_index(index)
        for scene_item in scene_items.values():
            scene_item.setActive(True)

    def deactivate(self, index):
        scene_items = self._items_from_index(index)
        for scene_item in scene_items.values():
            scene_item.setActive(False)


class VolumeTreeItemModel(TreeItemModel):
    def __init__(self, data: ep.EyeVolume, parent):
        super().__init__(parent=parent)

        self._data = data
        self._current_slice = 0

        if not "AreaSettings" in self._data.meta:
            self._data.meta["AreaSettings"] = {
                "visible": True,
                "z_value": 0,
                "name": "Areas",
            }
        if not "LayerSettings" in self._data.meta:
            self._data.meta["LayerSettings"] = {
                "visible": True,
                "z_value": 0,
                "name": "Layers",
            }

        self._init_model()
        self._annotations = DynamicDefaultDict(
            lambda index: self._get_annotations(index)
        )

        self.annotation_items = collections.defaultdict(lambda: {})
        self.dataChanged.connect(self.sync_annotations_to_tab)

        self._scenes = {}

    def duplicate_volume(self, index: QtCore.QModelIndex):
        annotation = index.internalPointer().annotation
        name = annotation.name
        data = annotation.data

        if index.parent() == self.layers_index:
            self.add_layer_annotation(name=f"Duplicate {name}", height_map=data)
        elif index.parent() == self.areas_index:
            self.add_voxel_annotation(name=f"Duplicate {name}", voxel_map=data)

    @property
    def scene(self) -> CustomGraphicsScene:
        if not self.current_slice in self._scenes:
            # Create GraphicsScene if not yet created
            scene = CustomGraphicsScene(
                parent=self, data=self._data[self.current_slice]
            )
            # Set Annotations managed by the ViewTab
            scene.addItem(self.annotations)
            self._scenes[self.current_slice] = scene

        return self._scenes[self.current_slice]

    def set_current_scene(self):
        if self.tool.paint_preview.scene() == self.scene:
            self.scene.removeItem(self.tool.paint_preview)
        if not self.scene.mouseGrabberItem() is None:
            self.tool.paint_preview.setParentItem(self.scene.mouseGrabberItem())

    def next_slice(self, current_tool):
        if self.current_slice < len(self._data) - 1:
            self.current_slice += 1
            if not self.scene.mouseGrabberItem() is None:
                current_tool.paint_preview.setParentItem(self.scene.mouseGrabberItem())

    def last_slice(self, current_tool):
        if self.current_slice > 0:
            self.current_slice -= 1
            if not self.scene.mouseGrabberItem() is None:
                current_tool.paint_preview.setParentItem(self.scene.mouseGrabberItem())

    def sync_annotations_to_tab(self):
        self.annotations.update()

    @property
    def annotations(self):
        return self._annotations[self.current_slice]

    @property
    def current_slice(self):
        return self._current_slice

    @current_slice.setter
    def current_slice(self, value: int):
        self._current_slice = value
        # self.annotations.sync_with_volume()
        self.annotations.update()

    def _init_model(self):
        self.tree_root = TreeItemGroup(parent=None, meta={"name": "root"})
        self.tree_areas = TreeItemGroup(meta=self._data.meta["AreaSettings"])
        self.tree_layers = TreeItemGroup(meta=self._data.meta["LayerSettings"])

        self.areas_index = QtCore.QPersistentModelIndex(
            self.createIndex(0, 0, self.tree_areas)
        )
        self.layers_index = QtCore.QPersistentModelIndex(
            self.createIndex(1, 0, self.tree_layers)
        )

        for vm in self._data.volume_maps.values():
            vm.meta = {
                **{"visible": True, "z_value": 0, "current_color": "FF0000"},
                **vm.meta,
            }
            self.tree_areas.appendChild(TreeItem(data=vm))
        for layer in self._data.layers.values():
            layer.meta = {
                **{"visible": True, "z_value": 0, "current_color": "FF0000"},
                **layer.meta,
            }
            self.tree_layers.appendChild(TreeItem(data=layer))

        self.tree_root.appendChild(self.tree_areas)
        self.tree_root.appendChild(self.tree_layers)

    def add_annotation(self):
        options = {
            "Layers": self.add_layer_annotation,
            "Areas": self.add_voxel_annotation,
        }
        dialog = AddAnnotationDialog(self, options)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            self.layoutChanged.emit()
            self.annotations.update()

    def _items_from_index(self, index):
        item = self.getItem(index)
        if type(item) is TreeItemGroup:
            return {}
        # Remove respective items from all Bscan scenes
        eyevolume_annotation = item.annotation
        return self.annotation_items[id(eyevolume_annotation)]

    def remove_annotation(self, index):
        """Remove annotation from eyevolume, all scenes and the overview"""
        if index.isValid():
            self.layoutAboutToBeChanged.emit()
            item = self.getItem(index)
            # Remove respective items from all Bscan scenes
            eyevolume_annotation = item.annotation
            scene_items = self.annotation_items.pop(id(eyevolume_annotation))
            for scene_item in scene_items.values():
                scene_item.scene().removeItem(scene_item)
                del scene_item
            # Remove from EyeVolume
            for annotations in [self._data._volume_maps, self._data._layers]:
                if eyevolume_annotation in annotations:
                    annotations.remove(eyevolume_annotation)

            self.removeRows(index.row(), 1, index.parent())
            self.layoutChanged.emit()
            self.annotations.update()

    def add_layer_annotation(self, name, height_map=None, color="FF0000"):
        # Add to EyeVolume
        layer = self._data.add_layer_annotation(height_map=height_map, name=name)
        layer.meta = {
            **{"visible": True, "z_value": 0, "current_color": color},
            **layer.meta,
        }

        # Add in ViewTab - update
        self.appendRow(TreeItem(data=layer), parent=self.layers_index)

        # Add to every slice annotations exist for
        for index in self._annotations:
            root_item = self._annotations[index]
            layers_item_group = [
                c for c in root_item.childItems() if c.meta["name"] == "Layers"
            ][0]
            item = LayerItem(data=layer, index=index, parent=layers_item_group)
            self.annotation_items[id(layer)][index] = item

    def add_voxel_annotation(self, name, voxel_map=None, color="FF0000"):
        # Add to EyeVolume
        voxel_map = self._data.add_voxel_annotation(voxel_map=voxel_map, name=name)
        voxel_map.meta = {
            **{"visible": True, "z_value": 0, "current_color": color},
            **voxel_map.meta,
        }

        # Add in ViewTab - update
        self.appendRow(TreeItem(data=voxel_map), parent=self.areas_index)

        # Add to every slice annotations exist for
        for index in self._annotations:
            root_item = self._annotations[index]
            areas_item_group = [
                c for c in root_item.childItems() if c.meta["name"] == "Areas"
            ][0]
            item = AreaItem(data=voxel_map, index=index, parent=areas_item_group)
            self.annotation_items[id(voxel_map)][index] = item

    def _get_annotations(self, index):
        root_item = ItemGroup(meta={})

        # Create AreaItems for the given index
        areas_item_group = ItemGroup(meta=self._data.meta["AreaSettings"])
        for name, volume_map in self._data.volume_maps.items():
            item = AreaItem(data=volume_map, index=index, parent=areas_item_group)
            # Save items organized by annotation for easier access
            self.annotation_items[id(volume_map)][index] = item

        # Create LayerItems for the given index
        layers_item_group = ItemGroup(meta=self._data.meta["LayerSettings"])
        for name, layer in self._data.layers.items():
            item = LayerItem(data=layer, index=index, parent=layers_item_group)
            # Save items organized by annotation for easier access
            self.annotation_items[id(layer)][index] = item

        areas_item_group.setParentItem(root_item)
        layers_item_group.setParentItem(root_item)
        return root_item


class EnfaceTreeItemModel(TreeItemModel):
    def __init__(self, data: ep.EyeEnface, parent):
        super().__init__(parent=parent)

        self._data = data

        if not "AreaSettings" in self._data.meta:
            self._data.meta["AreaSettings"] = {
                "visible": True,
                "z_value": 0,
                "name": "Areas",
            }

        self.annotation_items = {}
        self._init_model()
        self.annotations = self._get_annotations()

        self.dataChanged.connect(self.sync_annotations_to_tab)

        self.scene = CustomGraphicsScene(parent=self, data=self._data)
        self.scene.addItem(self.annotations)

    def sync_annotations_to_tab(self):
        # self.annotations.sync_with_volume()
        self.annotations.update()

    def _init_model(self):
        self.tree_root = TreeItemGroup(parent=None, meta={"name": "root"})
        self.tree_areas = TreeItemGroup(meta=self._data.meta["AreaSettings"])

        self.areas_index = QtCore.QPersistentModelIndex(
            self.createIndex(0, 0, self.tree_areas)
        )

        for vm in self._data.area_maps.values():
            vm.meta = {
                **{"visible": True, "z_value": 0, "current_color": "FF0000"},
                **vm.meta,
            }
            self.tree_areas.appendChild(TreeItem(data=vm))

        self.tree_root.appendChild(self.tree_areas)

    def _get_annotations(self):
        root_item = ItemGroup(meta={})

        # Create AreaItems for the given index
        areas_item_group = ItemGroup(meta=self._data.meta["AreaSettings"])
        for area_map in self._data._area_maps:
            item = AreaItem(
                data=area_map,
                index=np.s_[...],
                parent=areas_item_group,
            )
            self.annotation_items[id(area_map)] = item

        areas_item_group.setParentItem(root_item)
        return root_item

    def _items_from_index(self, index):
        item = self.getItem(index)
        if type(item) is TreeItemGroup:
            return {}
        # Remove respective items from all Bscan scenes
        voxel_annotation = item.annotation
        return {0: self.annotation_items[id(voxel_annotation)]}

    def add_annotation(self):
        dialog = AddAnnotationDialog(self, {"Areas": self.add_area_annotation})
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            self.layoutChanged.emit()
            self.annotations.update()

    def add_area_annotation(self, name, color="FF0000"):
        # Add to EyeVolume
        area_map = self._data.add_area_annotation(name=name)
        area_map.meta = {
            **{"visible": True, "z_value": 0, "current_color": color},
            **area_map.meta,
        }

        # Add in ViewTab - update
        self.appendRow(TreeItem(data=area_map), parent=self.areas_index)

        # Add to every slice annotations exist for
        root_item = self.annotations
        areas_item_group = [
            c for c in root_item.childItems() if c.meta["name"] == "Areas"
        ][0]
        item = AreaItem(data=area_map, parent=areas_item_group)
        self.annotation_items[id(area_map)] = item

    def remove_annotation(self, index):
        """Remove annotation from eyevolume, all scenes and the overview"""
        if index.isValid():
            self.layoutAboutToBeChanged.emit()
            item = self.getItem(index)
            # Remove respective item from scene
            area_annotation = item.annotation
            scene_item = self.annotation_items.pop(id(area_annotation))
            scene_item.scene().removeItem(scene_item)
            del scene_item

            # Remove from EyeEnface
            self._data._area_maps.remove(area_annotation)

            self.removeRows(index.row(), 1, index.parent())
            self.layoutChanged.emit()
            self.annotations.update()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_text_window.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore


class Ui_TextWindow(object):
    def setupUi(self, TextWindow):
        if not TextWindow.objectName():
            TextWindow.setObjectName(u"TextWindow")
        TextWindow.resize(400, 300)
        self.horizontalLayout = QHBoxLayout(TextWindow)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.textEdit = QTextEdit(TextWindow)
        self.textEdit.setObjectName(u"textEdit")
        self.textEdit.setReadOnly(True)

        self.horizontalLayout.addWidget(self.textEdit)

        self.retranslateUi(TextWindow)

        QMetaObject.connectSlotsByName(TextWindow)

    # setupUi

    def retranslateUi(self, TextWindow):
        TextWindow.setWindowTitle(
            QCoreApplication.translate("TextWindow", u"Help", None)
        )
        self.textEdit.setHtml(
            QCoreApplication.translate(
                "TextWindow",
                u'<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">\n'
                '<html><head><meta name="qrichtext" content="1" /><style type="text/css">\n'
                "p, li { white-space: pre-wrap; }\n"
                "</style></head><body style=\" font-family:'Ubuntu'; font-size:11pt; font-weight:400; font-style:normal;\">\n"
                '<p style="-qt-paragraph-type:empty; margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;"><br /></p></body></html>',
                None,
            )
        )

    # retranslateUi



import qimage2ndarray
from PySide6 import QtGui, QtWidgets


def array2qgraphicspixmapitem(image):
    return QtWidgets.QGraphicsPixmapItem(
        QtGui.QPixmap().fromImage(qimage2ndarray.array2qimage(image))
    )


def qgraphicspixmapitem2array(pixmapitem):
    return qimage2ndarray.rgb_view(pixmapitem.pixmap().toImage())


def qgraphicspixmap2array(pixmap):
    return qimage2ndarray.rgb_view(pixmap.toImage())



import logging

import eyepy as ep
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtGui import QCloseEvent, QUndoStack
from PySide6.QtWidgets import QMessageBox, QWidget

from eyelab.commands import get_undo_stack
from eyelab.views.ui.ui_workspace_view import Ui_WorkspaceView

logger = logging.getLogger("eyelab.workspace")


class Workspace(QWidget, Ui_WorkspaceView):
    def __init__(self, parent=None):
        """Holds DataView and Toolbox."""
        logger.debug("Workspace: __init__")
        super().__init__(parent)
        self.setupUi(self)

        self.undo_stack = get_undo_stack("main", self)
        self.graphic_views = [
            self.data_view.graphicsViewLocalizer,
            self.data_view.graphicsViewVolume,
        ]

        self._tools = {}

        self.ctrl_key_actions = {
            QtCore.Qt.Key_X: self.toggle_linked_navigation,
            # QtCore.Qt.Key_A: self.toggle_all_annotations,
            # QtCore.Qt.Key_1: self.select_tool_0,
            # QtCore.Qt.Key_2: self.select_tool_1
        }

        self.linked_navigation = False
        self.annotations_visible = True

        self.data = None

    def closeEvent(self, event: QCloseEvent) -> None:
        # Save all unchanged annotations
        if self.undo_stack.isClean():
            super().closeEvent(event)
            return

        answer = QMessageBox.question(
            self,
            None,
            "You have unsaved changes. Save before closing?",
            QMessageBox.Save | QMessageBox.Discard | QMessageBox.Cancel,
        )
        if answer & QMessageBox.Save:
            # the user chose "Save"
            self.actionSaveAnnotations.trigger()
        elif answer & QMessageBox.Cancel:
            # the user canceled
            event.ignore()

    def set_data(self, data: ep.EyeVolume):
        logger.debug("Workspace: set_data")
        self.data = data
        self.data_view.set_data(data)
        self.undo_stack.clear()

        self.layer_overview.clear()
        self.layer_overview.addTab(self.data_view.graphicsViewVolume.view_tab, "Volume")
        self.layer_overview.addTab(
            self.data_view.graphicsViewLocalizer.view_tab, "Localizer"
        )
        self.layer_overview.repaint()

    @property
    def scenes(self):
        return [view.scene() for view in self.graphic_views]

    def toggle_linked_navigation(self):
        if self.linked_navigation:
            self.linked_navigation = False
            for view in self.graphic_views:
                view.unlink_navigation()

        else:
            self.linked_navigation = True
            for view in self.graphic_views:
                view.link_navigation()

    def toggle_all_annotations(self):
        for i in range(self.layer_overview.count()):
            self.layer_overview.widget(i).model.toggle_annotations()

    def select_tool_0(self):
        self.select_tool(0)

    def select_tool_1(self):
        self.select_tool(1)

    def select_tool(self, number):
        for view in self.graphic_views:
            if view.isEnabled():
                layout = view.view_tab.toolsWidget.layout()
                widget = layout.itemAt(number).widget()
                if not widget is None:
                    widget.click()

    def uncheck_buttons(self):
        for button in [self.penButton, self.navigationButton]:
            button.setChecked(False)

    def toogle_scroll_bars(self):
        [view.toggle_scroll_bars() for view in self.graphic_views]

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ControlModifier:
            if event.key() in self.ctrl_key_actions:
                self.ctrl_key_actions[event.key()]()
                event.accept()
        super().keyPressEvent(event)



from eyepy import EyeEnface
from PySide6 import QtGui, QtWidgets
from PySide6.QtCore import QPointF, Qt

from eyelab.models.viewtab import EnfaceTab
from eyelab.views.graphicsview import CustomGraphicsView


class EnfaceView(CustomGraphicsView):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)

        self.setContextMenuPolicy(Qt.ActionsContextMenu)

        bscan_overlay_action = QtGui.QAction()
        bscan_overlay_action.setCheckable(True)
        bscan_overlay_action.setChecked(True)
        bscan_overlay_action.setText("B-scan positions")
        bscan_overlay_action.triggered.connect(self.test)
        self.addAction(bscan_overlay_action)

    def test(self):
        print("test")

    def set_data(self, data: EyeEnface, name: str):
        self.data = data
        self.view_tab = EnfaceTab(self.data)
        self.setScene(self.view_tab.model.scene)
        self.zoomToFit()

    def map_from_sender(self, pos, sender):
        pass
        # tform = self.get_tform(sender)
        # result = tform((pos.x(), pos.y()))[0]
        # return QPointF(*result)

    def map_to_sender(self, pos, sender):
        tform = self.get_tform(sender)
        result = tform.inverse((pos.x(), pos.y()))[0]
        return QPointF(*result)

    def set_fake_cursor(self, pos, sender=None):
        pos = QPointF(pos.x(), pos.y())
        if not sender is None and sender != self:
            pos = self.map_from_sender(pos, sender)
        if self.linked_navigation:
            self.centerOn(pos)
        self.scene().fake_cursor.setPos(pos)
        self.scene().fake_cursor.show()
        self.viewport().update()

    def wheelEvent(self, event):
        if event.modifiers() == (Qt.ControlModifier):
            event.accept()
        else:
            super().wheelEvent(event)

    def mouseMoveEvent(self, event):
        super().mouseMoveEvent(event)
        scene_pos = self.mapToScene(event.pos())
        self.cursorPosChanged.emit(scene_pos, self)

    def get_tform(self, other_view):
        id_pair = (self.image_id, other_view.scene().image_id)
        print("here")
        if not id_pair in self._tforms:
            print("here_to")
            tmodel = "similarity"
            self._tforms[id_pair] = get_transformation(*id_pair, tmodel)
        return self._tforms[id_pair]

    # def scrollContentsBy(self, dx: int, dy: int) -> None:
    #    super().scrollContentsBy(dx, dy)
    #    self.viewChanged.emit(self)

    # def match_viewport(self, view: QtWidgets.QGraphicsView):
    #    print("match_viewport")
    #    if self.linked_navigation:
    #        rect = self.map_rect_from_sender(view.rect(), view)
    #        print("linked is true")
    #        self.fitInView(rect)

    # def map_rect_from_sender(self, rect: QtCore.QRect, sender):
    #    tform = self.get_tform(sender)
    #    print(tform)
    #    top_left = tform((rect.topLeft().x(), rect.topLeft().y()))[0]
    #    bot_right = tform((rect.bottomRight().x(), rect.bottomRight().y()))[0]
    #    return QtCore.QRectF(QtCore.QPointF(*top_left), QtCore.QPointF(*bot_right))



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_workspace_view.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore

from eyelab.views.data_view import DataView

from . import resources_rc


class Ui_WorkspaceView(object):
    def setupUi(self, WorkspaceView):
        if not WorkspaceView.objectName():
            WorkspaceView.setObjectName(u"WorkspaceView")
        WorkspaceView.resize(856, 489)
        self.horizontalLayout = QHBoxLayout(WorkspaceView)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.data_view = DataView(WorkspaceView)
        self.data_view.setObjectName(u"data_view")
        sizePolicy = QSizePolicy(
            QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding
        )
        sizePolicy.setHorizontalStretch(200)
        sizePolicy.setVerticalStretch(200)
        sizePolicy.setHeightForWidth(self.data_view.sizePolicy().hasHeightForWidth())
        self.data_view.setSizePolicy(sizePolicy)
        self.data_view.setAutoFillBackground(False)
        self.data_view.setStyleSheet(u"")

        self.horizontalLayout.addWidget(self.data_view)

        self.layer_overview = QTabWidget(WorkspaceView)
        self.layer_overview.setObjectName(u"layer_overview")
        sizePolicy1 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(
            self.layer_overview.sizePolicy().hasHeightForWidth()
        )
        self.layer_overview.setSizePolicy(sizePolicy1)
        self.layer_overview.setMinimumSize(QSize(170, 250))
        self.layer_overview.setElideMode(Qt.ElideNone)

        self.horizontalLayout.addWidget(self.layer_overview)

        self.retranslateUi(WorkspaceView)

        self.layer_overview.setCurrentIndex(-1)

        QMetaObject.connectSlotsByName(WorkspaceView)

    # setupUi

    def retranslateUi(self, WorkspaceView):
        WorkspaceView.setWindowTitle(
            QCoreApplication.translate("WorkspaceView", u"Workspace", None)
        )

    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_spline_options.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore


class Ui_splineOptions(object):
    def setupUi(self, splineOptions):
        if not splineOptions.objectName():
            splineOptions.setObjectName(u"splineOptions")
        splineOptions.resize(139, 109)
        self.verticalLayout = QVBoxLayout(splineOptions)
        self.verticalLayout.setSpacing(2)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(2, 2, 2, 2)
        self.label = QLabel(splineOptions)
        self.label.setObjectName(u"label")
        font = QFont()
        font.setPointSize(8)
        font.setBold(True)
        self.label.setFont(font)

        self.verticalLayout.addWidget(self.label)

        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setSpacing(2)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(5, -1, -1, -1)
        self.showCheckBox = QCheckBox(splineOptions)
        self.showCheckBox.setObjectName(u"showCheckBox")
        font1 = QFont()
        font1.setPointSize(8)
        self.showCheckBox.setFont(font1)
        self.showCheckBox.setChecked(True)

        self.verticalLayout_2.addWidget(self.showCheckBox)

        self.strengthCheckBox = QCheckBox(splineOptions)
        self.strengthCheckBox.setObjectName(u"strengthCheckBox")
        self.strengthCheckBox.setFont(font1)
        self.strengthCheckBox.setChecked(True)

        self.verticalLayout_2.addWidget(self.strengthCheckBox)

        self.slopeCheckBox = QCheckBox(splineOptions)
        self.slopeCheckBox.setObjectName(u"slopeCheckBox")
        self.slopeCheckBox.setFont(font1)
        self.slopeCheckBox.setChecked(True)

        self.verticalLayout_2.addWidget(self.slopeCheckBox)

        self.neighbourCheckBox = QCheckBox(splineOptions)
        self.neighbourCheckBox.setObjectName(u"neighbourCheckBox")
        self.neighbourCheckBox.setFont(font1)
        self.neighbourCheckBox.setChecked(True)

        self.verticalLayout_2.addWidget(self.neighbourCheckBox)

        self.verticalLayout.addLayout(self.verticalLayout_2)

        self.verticalSpacer = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self.verticalLayout.addItem(self.verticalSpacer)

        self.retranslateUi(splineOptions)

        QMetaObject.connectSlotsByName(splineOptions)

    # setupUi

    def retranslateUi(self, splineOptions):
        splineOptions.setWindowTitle("")
        self.label.setText(
            QCoreApplication.translate("splineOptions", u"Control points:", None)
        )
        # if QT_CONFIG(tooltip)
        self.showCheckBox.setToolTip(
            QCoreApplication.translate(
                "splineOptions", u"Hide the blue control points", None
            )
        )
        # endif // QT_CONFIG(tooltip)
        self.showCheckBox.setText(
            QCoreApplication.translate("splineOptions", u"Show", None)
        )
        # if QT_CONFIG(tooltip)
        self.strengthCheckBox.setToolTip(
            QCoreApplication.translate(
                "splineOptions",
                u"Optimize control point position when moving a knot",
                None,
            )
        )
        # endif // QT_CONFIG(tooltip)
        self.strengthCheckBox.setText(
            QCoreApplication.translate("splineOptions", u"Optimize strength", None)
        )
        self.slopeCheckBox.setText(
            QCoreApplication.translate("splineOptions", u"Optimize slope", None)
        )
        self.neighbourCheckBox.setText(
            QCoreApplication.translate("splineOptions", u"Optimize neighbours", None)
        )

    # retranslateUi



from PySide6 import QtCore, QtGui, QtWidgets


class FillSameWidget(QtWidgets.QWidget, Ui_fillSameOptions):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        self.sizeSlider.valueChanged.connect(self.set_label)
        self.set_label()

    def set_label(self):
        self.sizeLabel.setText(f"Size: {self.sizeSlider.value()}")


class PaintPreview(QtWidgets.QGraphicsPixmapItem):
    def __init__(self, tool, parent=None):
        super().__init__(parent)
        self.tool = tool
        self.settings_widget = tool.options_widget
        self.settings_widget.sizeSlider.valueChanged.connect(self.set_preview)
        self.set_preview()

    def set_preview(self):
        diameter = self.settings_widget.sizeSlider.value()
        pm = QtGui.QPixmap(diameter + 1, diameter + 1)
        pm.fill(QtCore.Qt.transparent)
        painter = QtGui.QPainter(pm)
        pen = QtGui.QPen(QtCore.Qt.red)
        painter.setPen(pen)
        painter.setBackgroundMode(QtCore.Qt.TransparentMode)
        painter.drawPixmap(0, 0, self.tool.mask)
        painter.end()

        self.setPixmap(pm)
        offset = self.settings_widget.sizeSlider.value() // 2
        self.setOffset(-offset, -offset)
        self.setOpacity(0.25)


class FillSame(object):
    def __init__(self):
        """ """
        self._masks = {}
        self.cursor = self.get_cursor()
        self.button = self.get_tool_button()
        self.hot_key = None
        self.options_widget = self.get_options_widget()
        self.paint_preview = self.get_paint_preview()

    def enable(self, view):
        pass

    def disable(self, view):
        pass

    @property
    def mask(self):
        diameter = self.options_widget.sizeSlider.value()
        if diameter not in self._masks:
            bm = QtGui.QBitmap(diameter + 1, diameter + 1)
            bm.fill(QtCore.Qt.color0)
            painter = QtGui.QPainter(bm)
            brush = QtGui.QBrush()
            brush.setStyle(QtCore.Qt.SolidPattern)
            painter.setBrush(brush)
            painter.drawEllipse(0, 0, diameter, diameter)
            painter.end()
            self._masks[diameter] = bm

        return self._masks[diameter]

    def get_paint_preview(self):
        return PaintPreview(self)

    def get_options_widget(self):
        widget = FillSameWidget()
        return widget

    def get_tool_button(self):
        button = QtWidgets.QToolButton()
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(":/icons/icons/baseline-edit-24px.svg"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        button.setIcon(icon)
        button.setIconSize(QtCore.QSize(24, 24))
        button.setCheckable(True)
        button.setObjectName("penButton")
        return button

    def get_cursor(self):
        return QtGui.QCursor(
            QtGui.QPixmap(":/cursors/cursors/pen_cursor.svg"), hotX=0, hotY=0
        )

    def get_painter(self, gitem):
        painter = QtGui.QPainter(gitem.qimage)
        color = QtGui.QColor(f"#{gitem.current_color}")
        brush = QtGui.QBrush()
        brush.setStyle(QtCore.Qt.SolidPattern)
        brush.setColor(color)
        painter.setPen(color)
        painter.setBrush(brush)
        return painter

    def draw(self, gitem, pos):
        radius = self.options_widget.sizeSlider.value() / 2
        painter = self.get_painter(gitem)
        painter.setBackgroundMode(QtCore.Qt.TransparentMode)
        pos = QtCore.QPointF(pos.x() - radius, pos.y() - radius)
        painter.drawPixmap(pos, self.mask)
        painter.end()

        gitem.update_pixmap()
        gitem.changed = True

    def erase(self, gitem, pos):
        radius = self.options_widget.sizeSlider.value() / 2
        painter = self.get_painter(gitem)
        painter.setCompositionMode(QtGui.QPainter.CompositionMode_Clear)
        pos = QtCore.QPointF(pos.x() - radius, pos.y() - radius)
        painter.drawPixmap(pos, self.mask)
        painter.end()

        gitem.update_pixmap()
        gitem.changed = True

    def mouse_move_handler(self, gitem: "TreeAreaItemDB", event):
        pos = gitem.mapToScene(event.pos()).toPoint()
        if event.buttons() & QtCore.Qt.LeftButton:
            self.draw(gitem, pos)
        elif event.buttons() & QtCore.Qt.RightButton:
            self.erase(gitem, pos)

    def mouse_press_handler(self, gitem, event):
        pos = gitem.mapToScene(event.pos()).toPoint()
        if event.buttons() & QtCore.Qt.LeftButton:
            self.draw(gitem, pos)
        elif event.buttons() & QtCore.Qt.RightButton:
            self.erase(gitem, pos)

    @staticmethod
    def mouse_release_handler(gitem, event):
        pass

    @staticmethod
    def key_press_handler(gitem, event):
        pass

    @staticmethod
    def key_release_handler(gitem, event):
        pass
